##
# File:    AlphaFoldModelProvider.py
# Author:  Dennis Piehl
# Date:    30-Sep-2021
#
# Updates:
#  15-Dec-2021 dwp Re-introduce use of FTP instead of HTTP for downloading files; proved to be significantly faster
#  11-Mar-2022 dwp Move reorganizing method to separate class to enable multiprocessing and to make it common to other provider classes;
#                  During reorganizing process, also rename the file to use an internal identifier instead of the original source filename;
#                  Change cache file to record new internal identifier, original filename, and accession URL for each model file;
#                  Add usage of config file object for specifying location for storing model files
#
# To Do:
# - Add check that converted files are consistent with mmCIF dictionaries
##
"""
Accessors for AlphaFold 3D Models (mmCIF).

"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

import datetime
import logging
import os.path
import time
from pathlib import Path
import copy
import glob
import re

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.FtpUtil import FtpUtil
from rcsb.utils.insilico3d.ModelReorganizer import ModelReorganizer

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger(__name__)


class AlphaFoldModelProvider:
    """Accessors for AlphaFold models (mmCIF)."""

    def __init__(self, cachePath, baseWorkPath=None, useCache=False, reload=True, **kwargs):
        """Initialize AlphaFoldModelProvider class.

        Args:
            cachePath (str): Path to directory where model files will be reorganized and stored permanently (also contains computed-model-cache.json file)  (i.e., 'computed-models'),
                             and will serve as the parent directory of where model files will be downloaded and where AF-specific cache file will sit (under 'work-dir/AlphaFold').
            useCache (bool, optional): Start from cache of already downloaded and/or reorganized model files. Defaults to False.
                                       When True, checks if last downloaded set of files is up-to-date and downloads any newly available models.
                                       When False (default), redownloads all model files.
            reload (bool, optional): Peform full reload (i.e., download/update) upon instantiation. Defaults to True.
        """
        # Use the same root cachePath for all types of insilico3D model sources, but with unique workPath names (sub-directory)
        self.__cachePath = cachePath  # Directory where model files will be reorganized and stored permanently (also contains computed-model-cache.json file) (i.e., 'computed-models')
        self.__baseWorkPath = baseWorkPath if baseWorkPath else self.__cachePath
        self.__workPath = os.path.join(self.__baseWorkPath, "work-dir", "AlphaFold")  # Directory where model files will be downloaded (also contains AF-specific cache file)
        self.__speciesDataCacheFile = os.path.join(self.__workPath, "model-download-cache.json")

        self.__ftpHost = kwargs.get("ftpHost", "ftp.ebi.ac.uk")
        self.__ftpDataPath = kwargs.get("ftpDataPath", "/pub/databases/alphafold/")

        self.__mU = MarshalUtil(workPath=self.__workPath)
        self.__fU = FileUtil(workPath=self.__workPath)
        self.__ftpU = FtpUtil(workPath=self.__workPath)

        if reload:
            self.__oD, self.__createdDate = self.__reload(useCache=useCache, **kwargs)

    def testCache(self, minCount=0):  # Increase minCount once we are consistently downloading more than one species data set
        if self.__oD and len(self.__oD) > minCount:
            return True
        else:
            return False

    def getCacheDirPath(self):
        return self.__workPath

    def reload(self, useCache, **kwargs):
        self.__oD, self.__createdDate = self.__reload(useCache=useCache, **kwargs)

    def __reload(self, **kwargs):
        """Reload cached list of species-specific AlphaFold model data files and check FTP server for updated data sets,
        or re-download latest versions of all species-specific model data sets from FTP server.

        Returns:
            oD (dict): dictionary of cached/downloaded species model data sets with general metadata about each data set (local data path, size, # files, ...)
            createdDate (str): timestamp in isoformat of when the data cache was last created
        """
        try:
            oD = None
            createdDate = None
            ok = False
            startDateTime = datetime.datetime.now().isoformat()
            useCache = kwargs.get("useCache", True)

            # alphaFoldLatestDataList = "https://ftp.ebi.ac.uk/pub/databases/alphafold/download_metadata.json"
            alphaFoldLatestDataList = "pub/databases/alphafold/download_metadata.json"
            alphaFoldRequestedSpeciesList = kwargs.get("alphaFoldRequestedSpeciesList", [])
            excludeArchiveFileRegexList = ["swissprot_pdb_v[0-9]+.tar"]
            excludeArchiveFileRegexListCombined = "(?:% s)" % "|".join(excludeArchiveFileRegexList)

            self.__ftpU.connect(self.__ftpHost)
            self.__fU.mkdir(self.__workPath)

            latestDataListDumpPath = os.path.join(self.__workPath, self.__fU.getFileName(alphaFoldLatestDataList))
            ok = self.__ftpU.get(alphaFoldLatestDataList, latestDataListDumpPath)
            lDL = self.__mU.doImport(latestDataListDumpPath, fmt="json")

            # Exclude undesired archives (defined in excludeArchiveFileRegexList)
            lDL = [s for s in lDL if not re.match(excludeArchiveFileRegexListCombined, s["archive_name"])]

            # If a specific list of species files was requested, only iterate over those
            if alphaFoldRequestedSpeciesList:
                alphaFoldArchiveDataList = [s for s in lDL if s.get("species", s.get("label", None)) in alphaFoldRequestedSpeciesList]
            else:
                alphaFoldArchiveDataList = lDL

            logger.info("useCache %r self.__speciesDataCacheFile %r", useCache, self.__speciesDataCacheFile)
            if useCache and self.__mU.exists(self.__speciesDataCacheFile):
                logger.info("Loading data cache, %s.", self.__speciesDataCacheFile)
                cacheD = self.__mU.doImport(self.__speciesDataCacheFile, fmt="json")
                createdDate = cacheD["created"]
                oD = cacheD["data"]
                cacheArchiveFileList = [sF for sF in oD]
                logger.info("Checking consistency of cached data with data available on FTP")
                for archiveD in alphaFoldArchiveDataList:
                    try:
                        speciesName = archiveD.get("species", archiveD.get("label", None))
                        archiveFile = archiveD["archive_name"]
                        archiveFileSize = int(archiveD["size_bytes"])
                        speciesNumModels = int(archiveD["num_predicted_structures"])
                        if speciesName not in cacheArchiveFileList:
                            logger.info("Species archive data for %s not found in local cache. Will re-fetch.", speciesName)
                            cacheD = self.fetchSpeciesArchive(archiveD, cacheD)
                            oD = cacheD["data"]
                            ok = self.__mU.doExport(self.__speciesDataCacheFile, cacheD, fmt="json", indent=4)
                            continue
                        #
                        # Check if cache was already reorganized
                        reorganized = archiveD.get("reorganized", False)
                        reorganizedBaseDir = archiveD.get("reorganizedBaseDir", None)
                        if reorganized and reorganizedBaseDir is not None:
                            if self.__mU.exists(reorganizedBaseDir):
                                logger.info("Species archive data for %s already reorganized to: %s", speciesName, reorganizedBaseDir)
                        #
                        cacheArchiveDir = oD[speciesName]["data_directory"]
                        cacheArchiveFileSize = oD[speciesName]["size_bytes"]
                        cacheSpeciesNumModels = oD[speciesName]["num_predicted_structures"]
                        if not os.path.exists(cacheArchiveDir) and not reorganized:
                            logger.warning("Species archive data directory for %s not found at cached path %s", archiveFile, cacheArchiveDir)
                        if cacheArchiveFileSize != archiveFileSize:
                            logger.warning("Species archive data file %s not up-to-date with file available on FTP server.", archiveFile)
                        if cacheSpeciesNumModels != speciesNumModels:
                            logger.warning(
                                "Missing some or all of the bundled model files for species archive %s as available on FTP server (found %d / %d model files)",
                                archiveFile,
                                cacheSpeciesNumModels,
                                speciesNumModels,
                            )
                    except Exception as e:
                        logger.exception("Failing on checking of cache data for %s from FTP server, with message:\n%s", archiveD["archive_name"], str(e))

            else:
                logger.info("Refetching all files from server.")
                cacheD = {}
                cacheD.update({"created": startDateTime, "data": {}})
                for archiveD in alphaFoldArchiveDataList:
                    cacheD = self.fetchSpeciesArchive(archiveD, cacheD)
                createdDate = cacheD["created"]
                oD = cacheD["data"]
                ok = self.__mU.doExport(self.__speciesDataCacheFile, cacheD, fmt="json", indent=4)
                logger.info("Export AlphaFold species model data (%d) status %r", len(oD), ok)

        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return oD, createdDate

    def fetchSpeciesArchive(self, archiveD, cacheD):
        try:
            sD = copy.deepcopy(archiveD)
            startTime = time.time()
            speciesName = sD.get("species", sD.get("label", None))
            archiveFile = sD["archive_name"]
            archiveFilePath = os.path.join(self.__ftpDataPath, "latest", archiveFile)
            speciesDataDumpDir = os.path.join(self.__workPath, speciesName.replace(" ", "_").replace("(", "").replace(")", ""))
            self.__fU.mkdir(speciesDataDumpDir)
            archiveFileDumpPath = os.path.join(speciesDataDumpDir, archiveFile)
            sD.update({"data_directory": speciesDataDumpDir, "archive_file_path": archiveFileDumpPath})

            logger.info("Fetching file %s from FTP server to local path %s", archiveFilePath, archiveFileDumpPath)
            ok = self.__ftpU.get(archiveFilePath, archiveFileDumpPath)
            ok = self.__fU.unbundleTarfile(archiveFileDumpPath, dirPath=speciesDataDumpDir)
            logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)

            logger.info("Clearing PDB files from extracted tar bundle...")
            for pdbDumpFile in Path(speciesDataDumpDir).glob("*.pdb.gz"):
                pdbDumpFile.unlink()

            if ok:
                cacheD["data"].update({speciesName: sD})
                self.__fU.remove(archiveFileDumpPath)

        except Exception as e:
            logger.exception("Failing on fetching and expansion of file %s from FTP server, with message:\n%s", archiveD["archive_name"], str(e))

        return cacheD

    def getArchiveDirList(self):
        archiveDirList = [self.__oD[k]["data_directory"] for k in self.__oD]

        return archiveDirList

    def getArchiveDataDict(self):
        archivDataDict = copy.deepcopy(self.__oD)
        return archivDataDict

    def getModelFileList(self, inputPathList=None):
        """Return a list of filepaths for all mmCIF models under the provided set of directories.

        Args:
            inputPathList (list): Optional (but recommended) list of paths to directories containing model mmCIF files to return.
                                  If not provided, method will retrieve all cached species data paths, so all model files from all species will be returned in a single list;
                                  thus, it's recommended to provide a list of specific species directory to break the returned model list down into more manageable parts.

        Returns:
            (list): list of model mmCIF file paths (only matches compressed ".cif.gz" files, to ensure only one file per model)
        """

        if not inputPathList:
            inputPathList = self.getArchiveDirList()

        modelFileList = []

        for modelDir in inputPathList:
            try:
                modelFiles = glob.glob(os.path.join(modelDir, "*.cif.gz"))
                modelFileList = [os.path.abspath(f) for f in modelFiles]
            except Exception as e:
                logger.exception("Failing with %s", str(e))

        return modelFileList

    def getSpeciesDataDownloadDate(self):
        return self.__createdDate

    def getBaseDataPath(self):
        return self.__workPath

    def getSpeciesDataCacheFilePath(self):
        return self.__speciesDataCacheFile

    def getModelReorganizer(self, cachePath=None, useCache=True, workPath=None, **kwargs):
        cachePath = cachePath if cachePath else self.__cachePath
        workPath = workPath if workPath else self.__workPath
        return ModelReorganizer(cachePath=cachePath, useCache=useCache, workPath=workPath, **kwargs)

    def getComputedModelsDataPath(self):
        return self.__cachePath

    def reorganizeModelFiles(self, cachePath=None, useCache=True, inputModelList=None, **kwargs):
        """Reorganize model files from organism-wide model listing to hashed directory structure and rename files
        to follow internal identifier naming convention.

        Args:
            cachePath (str): Path to cache directory.
            inputModelList (list, optional): List of input model filepaths to reorganize; defaults to all models for all species model sets.
            **kwargs (optional):
                numProc (int): number of processes to use; default 2.
                chunkSize (int): incremental chunk size used for distribute work processes; default 20.
                keepSource (bool): whether to copy files to new directory (instead of moving them); default False.
                cacheFilePath (str): full filepath and name for cache file containing a dictionary of all reorganized models.

        Returns:
            (bool): True if successful; False otherwise.
        """
        try:
            ok = False
            #
            mR = self.getModelReorganizer(cachePath=cachePath, useCache=useCache, **kwargs)
            #
            if inputModelList:  # Only reorganize given list of model files
                ok = mR.reorganize(inputModelList=inputModelList, modelSource="AlphaFold", destBaseDir=self.__cachePath, useCache=useCache)
                if not ok:
                    logger.error("Reorganization of model files failed for inputModelList starting with item, %s", inputModelList[0])
            #
            else:  # Reorganize ALL model files for ALL available species model sets
                cacheD = self.__mU.doImport(self.__speciesDataCacheFile, fmt="json")
                archiveDataD = self.getArchiveDataDict()
                for species, archiveD in archiveDataD.items():
                    archiveDir = archiveD["data_directory"]
                    # First check if cache was already reorganized
                    if cacheD["data"][species]["data_directory"] == archiveDir:
                        reorganized = archiveD.get("reorganized", False)
                        reorganizedBaseDir = archiveD.get("reorganizedBaseDir", None)
                        if reorganized and reorganizedBaseDir is not None:
                            if self.__mU.exists(reorganizedBaseDir) and reorganizedBaseDir == self.__cachePath:
                                logger.info("Species archive data for %s already reorganized to: %s", species, reorganizedBaseDir)
                                ok = True
                                continue
                    # Proceed with reorganization
                    inputModelList = self.getModelFileList(inputPathList=[archiveDir])
                    ok = mR.reorganize(inputModelList=inputModelList, modelSource="AlphaFold", destBaseDir=self.__cachePath, useCache=useCache)
                    if not ok:
                        logger.error("Reorganization of model files failed for species archive %s", archiveDir)
                        break
                    # Update the cache file to indicate that the given species archive has been reorganized
                    if cacheD["data"][species]["data_directory"] == archiveDir:
                        cacheD["data"][species].update({"reorganized": True, "reorganizedBaseDir": self.__cachePath})
                        logger.info("Reorganization of model files complete for species, %s, from archive %s", species, archiveDir)
                    ok = self.__mU.doExport(self.__speciesDataCacheFile, cacheD, fmt="json", indent=4)
        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False
        #
        return ok



##
# File:    ModelReorganizer.py
# Author:  Dennis Piehl
# Date:    10-Mar-2022
#
# Updates:
#   28-Jun-2022  dwp Add __rebuildEntryIds() method to replace (or remove) source entry identifiers with internal IDs, to enable operability with RCSB.org tools.
#                    Note that these modified data files will NOT be publicly served or available, and proper attribution using the source entry IDs will be given.
#   29-Jun-2022  dwp Add both source and internal IDs to database_2 category in internal mmCIF file to maintain a reference and mapping to the source DB
#
# To Do:
# - pylint: disable=fixme
# - Add mkdssp calculation
##

"""
Multiprocessor worker methods for processing and organizing computed model files.

"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os.path
import copy
from datetime import datetime
import pytz

from mmcif.api.DataCategory import DataCategory
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger(__name__)


class ModelWorker(object):
    """A skeleton class that implements the interface expected by the multiprocessing
    for working on model files.
    """

    def __init__(self, **kwargs):
        _ = kwargs
        self.__workPath = kwargs.get("workPath", None)
        self.__fU = FileUtil(workPath=self.__workPath)
        self.__mU = MarshalUtil(workPath=self.__workPath)

    def reorganize(self, dataList, procName, optionsD, workingDir):
        """Enumerate and reorganize and rename model files.

        Args:
            dataList (list): list of model files to work on
            procName (str): worker process name
            optionsD (dict): dictionary of additional options that worker can access
            workingDir (str): path to working directory

        Returns:
            successList (list): list of input data items that were successfully processed
            retList (list): list of processed results
            diagList (list): list of unique diagnostics
        """
        _ = workingDir
        _ = optionsD
        successList = []
        failList = []
        retList = []
        diagList = []

        try:
            modelSourcePrefix = optionsD.get("modelSourcePrefix")  # e.g., "AF" or "MA"
            modelSourceDbMap = optionsD.get("modelSourceDbMap")
            destBaseDir = optionsD.get("destBaseDir")  # base path for all computed models (i.e., "computed-models"); Or will be root path at HTTP endpoint
            keepSource = optionsD.get("keepSource", False)  # whether to copy files over (instead of moving them)
            reorganizeDate = optionsD.get("reorganizeDate", False)  # whether to copy files over (instead of moving them)
            #
            for modelFileIn in dataList:
                modelD = {}
                success = False
                modelFileOut = None
                modelFileNameIn = self.__fU.getFileName(modelFileIn)
                modelSourceDb = modelSourceDbMap[modelSourcePrefix]
                #
                containerList = self.__mU.doImport(modelFileIn, fmt="mmcif")
                if len(containerList) > 1:
                    # Expecting all computed models to have one container per file. When this becomes no longer the case, update this to handle it accordingly.
                    logger.error("Skipping - model file %s has more than one container (%d)", modelFileNameIn, len(containerList))
                    continue
                #
                dataContainer = containerList[0]
                #
                # Create internal model ID using entry.id and strip away all punctuation and make ALL CAPS
                tObj = dataContainer.getObj("entry")
                sourceModelEntryId = tObj.getValue("id", 0)
                modelEntryId = "".join(char for char in sourceModelEntryId if char.isalnum()).upper()
                internalModelId = modelSourcePrefix + "_" + modelEntryId
                #
                dataContainer = self.__rebuildEntryIds(
                    dataContainer=dataContainer,
                    sourceModelEntryId=sourceModelEntryId,
                    sourceModelDb=modelSourceDb,
                    internalModelId=internalModelId
                )
                #
                # Get the revision date if it exists
                if dataContainer.exists("pdbx_audit_revision_history"):
                    lastModifiedDate = dataContainer.getObj("pdbx_audit_revision_history").getValue("revision_date", -1)
                    lastModifiedDate = datetime.strptime(lastModifiedDate, '%Y-%m-%d').replace(microsecond=0).replace(tzinfo=pytz.UTC).isoformat()
                else:
                    lastModifiedDate = reorganizeDate
                #
                # Insert default deposited pdbx_assembly information into CIF
                dataContainer = self.__addDepositedAssembly(dataContainer=dataContainer)
                #
                # Gzip the original file if not already (as the case for ModelArchive model files)
                if modelFileIn.endswith(".gz"):
                    modelFileInGzip = modelFileIn
                else:
                    modelFileInGzip = modelFileIn + ".gz"
                    logger.debug("Compressing model file %s --> %s", modelFileIn, modelFileInGzip)
                    ok = self.__fU.compress(modelFileIn, modelFileInGzip)
                    if not ok:
                        logger.warning("Failed to gzip input model file: %s", modelFileIn)
                #
                internalModelName = internalModelId + ".cif.gz"
                #
                # Use last six to last two characters for second-level hashed directory
                firstDir, secondDir = modelEntryId[-6:-4], modelEntryId[-4:-2]
                modelPathFromPrefixDir = os.path.join(modelSourcePrefix, firstDir, secondDir, internalModelName)
                destModelDir = os.path.join(destBaseDir, modelSourcePrefix, firstDir, secondDir)
                if not self.__fU.exists(destModelDir):
                    self.__fU.mkdir(destModelDir)
                modelFileOut = os.path.join(destModelDir, internalModelName)
                modelFileOutUnzip = modelFileOut.split(".gz")[0]
                #
                sourceModelUrl = self.__getSourceUrl(modelSourcePrefix, modelFileNameIn, sourceModelEntryId)
                #
                modelD["modelId"] = internalModelId
                modelD["modelPath"] = modelPathFromPrefixDir  # Starts at prefix (e.g., "AF/XJ/E6/AF_AFA0A385XJE6F1.cif.gz"); needed like this by RepositoryProvider
                modelD["sourceId"] = sourceModelEntryId
                modelD["sourceDb"] = modelSourceDb
                modelD["sourceModelFileName"] = modelFileNameIn
                if sourceModelUrl:
                    modelD["sourceModelUrl"] = sourceModelUrl
                modelD["lastModifiedDate"] = lastModifiedDate
                #
                try:
                    self.__mU.doExport(modelFileOutUnzip, containerList, fmt="mmcif")
                    self.__fU.compress(modelFileOutUnzip, modelFileOut)
                    self.__mU.remove(modelFileOutUnzip)
                    if not keepSource:
                        self.__mU.remove(modelFileInGzip)  # Remove original file
                        if self.__mU.exists(modelFileIn):   # Remove unzipped file too if it exists
                            self.__mU.remove(modelFileIn)
                    success = True
                    successList.append(modelFileIn)
                except Exception as e:
                    logger.debug("Failing to reorganize %s --> %s, with %s", modelFileIn, modelFileOut, str(e))
                #
                retList.append((modelFileIn, modelD, success))
                #
            failList = sorted(set(dataList) - set(successList))
            if failList:
                logger.info("%s returns %d definitions with failures: %r", procName, len(failList), failList)
            #
            logger.debug("%s processed %d/%d models, failures %d", procName, len(retList), len(dataList), len(failList))
        except Exception as e:
            logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))

        return successList, retList, diagList

    def __getSourceUrl(self, modelSourcePrefix, sourceModelFileName, sourceModelEntryId):
        """Construct model accession URL for each model source.

        Args:
            modelSourcePrefix (str): Prefix for given model source (e.g., "AF" or "MA").
            sourceModelFileName (str): filename of model as provided by the source.
            sourceModelEntryId (str): model entry.id value as in the provided mmCIF file

        Returns:
            (str): Accession URL for the specified model file and source.
                    Note that file-specific downloads aren't gzipped, unlike model files in species tarball.
                    E.g., "https://alphafold.ebi.ac.uk/files/AF-Q9RQP8-F1-model_v2.cif"
        """
        sourceModelUrl = None
        try:
            if modelSourcePrefix == "AF":
                if not sourceModelEntryId.upper().endswith("F1"):
                    return None
                modelFileNameInUrl = sourceModelFileName.split(".gz")[0]
                sourceModelUrl = os.path.join("https://alphafold.ebi.ac.uk/files/", modelFileNameInUrl)
            elif modelSourcePrefix == "MB":
                modbaseInternalId = sourceModelEntryId.split("model_")[-1]
                # sourceModelUrl = "https://salilab.org/modbase/searchbyid?modelID=" + modbaseInternalId + "&displaymode=moddetail"  # Directs to entry page
                sourceModelUrl = "https://salilab.org/modbase/retrieve/modbase/?modelID=" + modbaseInternalId + "&format=mmcif"  # Directs to mmCIF file displayed in web browser
                # E.g.: https://salilab.org/modbase/retrieve/modbase/?modelID=ecac68b60ee6877ccde36af05cdeac58&format=mmcif
            elif modelSourcePrefix == "MA":
                modelFileNameInUrl = sourceModelFileName.split(".cif")[0]
                sourceModelUrl = "https://www.modelarchive.org/api/projects/" + modelFileNameInUrl + "?type=basic__model_file_name"
            else:
                logger.error("URL generation process not ready yet for %s", modelSourcePrefix)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return sourceModelUrl

    def __rebuildEntryIds(self, dataContainer, sourceModelEntryId, sourceModelDb, internalModelId):
        """Replace or remove occurrences of source entry ID with internal ID

        Change the entry ID to internal ID for the following items:
        - Top-level "data_" block name (e.g., data_ma-bak-cepc-1100)
        - _entry.id
        - _struct.entry_id
        - _atom_sites.entry_id
        - _struct_ref_seq.pdbx_PDB_id_code
        - _pdbx_database_status.entry_id

        Remove the following categories (or attributes) from the mmCIF completely:
        - _entry.ma_collection_id
        - _database_2.*  (entire category loop)  - will repopulate afterwards
        - _ma_entry_associated_files.*  (entire category loop)

        And ADD the following:
        - _database_2.*  - add references to both sourceDB ID and internal RCSB ID

        Args:
            dataContainer (object): mmcif.api.DataContainer object instance
            sourceModelEntryId (str): source entry ID
            sourceModelDb (str): source DB of model (must adhere to database_2.database_id enumerations)
            internalModelId (str): internal entry ID

        Returns:
            dataContainer: updated dataContainer object
        """
        # Replace "data_" block name
        dataContainer.setName(internalModelId)

        # Replace entry.id
        tObj = dataContainer.getObj("entry")
        tObj.setValue(internalModelId, "id", 0)

        # Remove entry.ma_collection_id attribute (if present)
        if tObj.hasAttribute("ma_collection_id"):
            tObj.removeAttribute("ma_collection_id")

        # Replace struct.entry_id
        if dataContainer.exists("struct"):
            tObj = dataContainer.getObj("struct")
            if tObj.hasAttribute("entry_id"):
                tObj.setValue(internalModelId, "entry_id", 0)

        # Replace atom_sites.entry_id
        if dataContainer.exists("atom_sites"):
            tObj = dataContainer.getObj("atom_sites")
            if tObj.hasAttribute("entry_id"):
                tObj.setValue(internalModelId, "entry_id", 0)

        # Replace struct_ref_seq.pdbx_PDB_id_code
        if dataContainer.exists("struct_ref_seq"):
            tObj = dataContainer.getObj("struct_ref_seq")
            if tObj.hasAttribute("pdbx_PDB_id_code"):
                tObj.setValue(internalModelId, "pdbx_PDB_id_code", 0)

        # If pdbx_database_status.* exisits, replace the source ID with internal ID, to match entry.id (parent-child relationship)
        if dataContainer.exists("pdbx_database_status"):
            tObj = dataContainer.getObj("pdbx_database_status")
            if tObj.hasAttribute("entry_id"):
                if tObj.getValue("entry_id", 0) == sourceModelEntryId:
                    tObj.setValue(internalModelId, "entry_id", 0)

        # Remove database_2.*  (entire category loop)
        if dataContainer.exists("database_2"):
            dataContainer.remove("database_2")

        # Remove ma_entry_associated_files.*  (entire category loop)
        if dataContainer.exists("ma_entry_associated_files"):
            dataContainer.remove("ma_entry_associated_files")

        # RE-ADD database_2.* with corrected references to both sourceDB ID and internal RCSB ID
        # (Note that the category will still be masked and NOT be loaded into ExDB, just saved to internal file - as of 29-Jun-2022 dwp)
        dataContainer.append(
            DataCategory("database_2", attributeNameList=["database_code", "database_id"], rowList=[[sourceModelEntryId, sourceModelDb], [internalModelId, "RCSB"]])
        )

        return dataContainer

    def __addDepositedAssembly(self, dataContainer):
        """Add the deposited coordinates as an additional separate assembly labeled as 'deposited'
        to categories, pdbx_struct_assembly and pdb_struct_assembly_gen.

        Method copied from rcsb.utils.dictionary.DictMethodAssemblyHelper.

        Args:
            dataContainer (object): mmcif.api.DataContainer object instance

        Returns:
            dataContainer: updated dataContainer object

        """
        if not dataContainer.exists("pdbx_struct_assembly"):
            dataContainer.append(
                DataCategory(
                    "pdbx_struct_assembly",
                    attributeNameList=["id", "details", "method_details", "oligomeric_details", "oligomeric_count", "rcsb_details", "rcsb_candidate_assembly"],
                )
            )
        if not dataContainer.exists("pdbx_struct_assembly_gen"):
            dataContainer.append(DataCategory("pdbx_struct_assembly_gen", attributeNameList=["assembly_id", "oper_expression", "asym_id_list", "ordinal"]))

        if not dataContainer.exists("pdbx_struct_oper_list"):
            row = [
                "1",
                "identity operation",
                "1_555",
                "x, y, z",
                "1.0000000000",
                "0.0000000000",
                "0.0000000000",
                "0.0000000000",
                "0.0000000000",
                "1.0000000000",
                "0.0000000000",
                "0.0000000000",
                "0.0000000000",
                "0.0000000000",
                "1.0000000000",
                "0.0000000000",
            ]
            atList = [
                "id",
                "type",
                "name",
                "symmetry_operation",
                "matrix[1][1]",
                "matrix[1][2]",
                "matrix[1][3]",
                "vector[1]",
                "matrix[2][1]",
                "matrix[2][2]",
                "matrix[2][3]",
                "vector[2]",
                "matrix[3][1]",
                "matrix[3][2]",
                "matrix[3][3]",
                "vector[3]",
            ]
            dataContainer.append(DataCategory("pdbx_struct_oper_list", attributeNameList=atList, rowList=[row]))
        #
        logger.debug("Add deposited assembly for %s", dataContainer.getName())
        cObj = dataContainer.getObj("struct_asym")
        asymIdL = cObj.getAttributeValueList("id")
        logger.debug("AsymIdL %r", asymIdL)
        #
        # Ordinal is added by subsequent attribure-level method.
        tObj = dataContainer.getObj("pdbx_struct_assembly_gen")
        rowIdx = tObj.getRowCount()
        tObj.setValue("deposited", "assembly_id", rowIdx)
        tObj.setValue("1", "oper_expression", rowIdx)
        tObj.setValue("1", "ordinal", rowIdx)
        tObj.setValue(",".join(asymIdL), "asym_id_list", rowIdx)
        #
        tObj = dataContainer.getObj("pdbx_struct_assembly")
        rowIdx = tObj.getRowCount()
        tObj.setValue("deposited", "id", rowIdx)
        tObj.setValue("deposited_coordinates", "details", rowIdx)
        tObj.setValue("Y", "rcsb_candidate_assembly", rowIdx)
        #
        for atName in ["oligomeric_details", "method_details", "oligomeric_count"]:
            if tObj.hasAttribute(atName):
                tObj.setValue("?", atName, rowIdx)
        #
        tObj.setValue(str(len(asymIdL)), "oligomeric_count", rowIdx)

        return dataContainer


class ModelReorganizer(object):
    """Generators and accessors for model files."""

    def __init__(self, cachePath=None, useCache=True, **kwargs):
        """Initialize ModelReorganizer object.

        Args:
            cachePath (str): directory path for storing cache file containing a dictionary of all reorganized models.
            useCache (bool): whether to use the existing data cache or re-run entire model reorganization process.
            cacheFile (str, optional): filename for cache file; default "computed-models-holdings.json.gz".
            cacheFilePath (str, optional): full filepath and name for cache file (will override "cachePath" and "cacheFile" if provided).
            numProc (int, optional): number of processes to use; default 2.
            chunkSize (int, optional): incremental chunk size used for distribute work processes; default 20.
            workPath (str, optional): directory path for workers to operate in; default is cachePath.
            keepSource (bool, optional): whether to copy model files to new directory instead of moving them; default False.
        """

        try:
            self.__cachePath = cachePath if cachePath else "."

            self.__numProc = kwargs.get("numProc", 2)
            self.__chunkSize = kwargs.get("chunkSize", 20)
            self.__workPath = kwargs.get("workPath", os.path.join(self.__cachePath, "work-dir"))
            self.__keepSource = kwargs.get("keepSource", False)

            self.__mU = MarshalUtil(workPath=self.__workPath)
            self.__fU = FileUtil(workPath=self.__workPath)

            self.__cacheFormat = kwargs.get("cacheFormat", "json")
            # self.__cacheFormat = kwargs.get("cacheFormat", "pickle")
            cacheExt = "pic" if self.__cacheFormat == "pickle" else "json"
            cacheFile = kwargs.get("cacheFile", "computed-models-holdings." + cacheExt + ".gz")
            cacheFilePath = kwargs.get("cacheFilePath", os.path.join(self.__cachePath, "holdings", cacheFile))
            if not cacheFilePath.lower().endswith(".gz"):
                logger.error("Holdings cache file must be gzipped, %s", cacheFilePath)
                raise ValueError("Error: Holdings cache file must be gzipped.")
            self.__cacheFilePath = cacheFilePath  # self.__cacheFilePath is full path to gzipped cache file
            self.__cacheFilePathUnzip = cacheFilePath[0:-3]

            logger.info("Reorganizing models using cachePath %s, cacheFilePath %s, cacheFilePathUnzip %s", self.__cachePath, self.__cacheFilePath, self.__cacheFilePathUnzip)

            self.__mD = self.__reload(cacheFilePath=self.__cacheFilePath, useCache=useCache)

        except Exception as e:
            logger.exception("Failing with %s", str(e))

    def testCache(self, minCount=20):
        try:
            if minCount == 0:
                return True
            if self.__mD and len(self.__mD) >= minCount:
                logger.info("Reorganized models in cache (%d)", len(self.__mD))
                return True
        except Exception:
            pass
        return False

    def reload(self, cacheFilePath, useCache=True):
        self.__mD = self.__reload(cacheFilePath, useCache=useCache)
        return len(self.__mD) > 0  # Returns True or False, if reload was successful or not

    def __reload(self, cacheFilePath, useCache=True):
        """Reload from the current cache file."""
        try:
            mD = {}
            logger.info("useCache %r cacheFilePath %r", useCache, cacheFilePath)
            if useCache and self.__mU.exists(cacheFilePath):
                if cacheFilePath != self.__cacheFilePath:
                    self.__cacheFilePath = cacheFilePath
                mD = self.__mU.doImport(cacheFilePath, fmt=self.__cacheFormat)
                logger.info("Reorganized models (%d) in cacheFilePath %r", len(mD), cacheFilePath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return mD

    def getCachePath(self):
        return self.__cachePath

    def getCacheFilePath(self):
        return self.__cacheFilePath

    def reorganize(self, inputModelList, modelSource, destBaseDir, useCache=True):
        """Move model files from organism-wide model listing to hashed directory structure and rename files
        to follow internal identifier naming convention.

        Args:
            inputModelList (list): List of input model filepaths to reorganize.
            modelSource (str): Source of model files ("AlphaFold", "ModBase", "ModelArchive", or "SwissModel")
            destBaseDir (str): Base destination directory into which to reorganize model files (e.g., "computed-models")

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            mD, failD = self.__reorganizeModels(
                inputModelList=inputModelList,
                modelSource=modelSource,
                destBaseDir=destBaseDir,
                numProc=self.__numProc,
                chunkSize=self.__chunkSize
            )
            if len(failD) > 0:
                logger.error("Failed to process %d model files.", len(failD))
            kwargs = {"indent": 4} if self.__cacheFormat == "json" else {"pickleProtocol": 4}
            if useCache:
                self.__mD = self.__reload(cacheFilePath=self.__cacheFilePath, useCache=useCache)
                for modelId, modelD in mD.items():
                    self.__mD.update({modelId: modelD})
            else:
                self.__mD = copy.deepcopy(mD)
            ok = self.__mU.doExport(self.__cacheFilePathUnzip, self.__mD, fmt=self.__cacheFormat, **kwargs)
            logger.info("Wrote %r status %r", self.__cacheFilePathUnzip, ok)
            if ok:
                ok2 = self.__fU.compress(self.__cacheFilePathUnzip, self.__cacheFilePath)
                logger.info("Compressed %r status %r", self.__cacheFilePath, ok2)
                if ok2:
                    ok3 = self.__fU.remove(self.__cacheFilePathUnzip)
                    logger.info("Removing uncompressed holdings cache file %s status %r", self.__cacheFilePathUnzip, ok3)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def __reorganizeModels(self, inputModelList, modelSource, destBaseDir, numProc=2, chunkSize=20):
        """Prepare multiprocessor queue and workers for generating input:output map for model files, to use in reorganizing
        them into a structured directory tree and naming them with internal identifiers.

        Args:
            inputModelList (list): List of input model filepaths to reorganize.
            modelSource (str): Source of model files ("AlphaFold", "ModBase", "ModelArchive", or "SwissModel")
            destBaseDir (str): Base destination directory into which to reorganize model files (e.g., "computed-models")
            numProc (int, optional): number of processes to use; default 2.
            chunkSize (int, optional): incremental chunk size used for distribute work processes; default 20.

        Returns:
            mD (dict): dictionary of successfully processed models, in the following structure:
                        { internalModelId_1 : {"sourceModelFileName": ..., "modelPath": ..., "sourceModelUrl": ...},
                          internalModelId_2 : {...}, ...}
                E.g:    { "AF_AFO25670F1" : {
                            "sourceModelFileName": "AF-O25670-F1-model_v2.cif.gz",
                            "modelPath": "./AF/56/70/AF_AFO25670F1.cif.gz",
                            "sourceModelUrl": "https://alphafold.ebi.ac.uk/files/AF-O25670-F1-model_v2.cif"},
                          "AF_AFO24939F1" : {...},
                        ...}
            failD (dict): dictionary of models for which processing failed and the associated output file path and source details
                          that were attempted, in the following structure:
                          {inputModelFilePath:  {"sourceModelFileName": ..., "modelPath": ..., "sourceModelUrl": ...}, ...}
        """
        mD = {}
        failD = {}
        tS = datetime.now().replace(microsecond=0).replace(tzinfo=pytz.UTC).isoformat()  # Desired format:  2022-04-15T12:00:00+00:00
        #
        logger.info("Starting with %d models, numProc %d at %r", len(inputModelList), numProc, tS)
        #
        # Create the base destination directory if it doesn't exist
        if not self.__fU.exists(destBaseDir):
            logger.info("Creating base destination directory for model file reorganization, %s", destBaseDir)
            self.__fU.mkdir(destBaseDir)
        #
        modelSourcePrefixD = {"AlphaFold": "AF", "ModBase": "MB", "ModelArchive": "MA", "SwissModelRepository": "SMR"}
        modelSourceDbMap = {"AF": "AlphaFoldDB", "MB": "MODBASE", "MA": "ModelArchive", "SMR": "SWISS-MODEL_REPOSITORY"}
        # Database name values correspond to _database_2.database_id enumerations (https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_database_2.database_id.html)
        modelSourcePrefix = modelSourcePrefixD[modelSource]
        #
        rWorker = ModelWorker(workPath=self.__workPath)
        mpu = MultiProcUtil(verbose=True)
        optD = {
            "modelSourcePrefix": modelSourcePrefix,
            "modelSourceDbMap": modelSourceDbMap,
            "destBaseDir": destBaseDir,
            "keepSource": self.__keepSource,
            "reorganizeDate": tS
        }
        mpu.setOptions(optD)
        mpu.set(workerObj=rWorker, workerMethod="reorganize")
        mpu.setWorkingDir(workingDir=self.__workPath)
        ok, failList, resultList, _ = mpu.runMulti(dataList=inputModelList, numProc=numProc, numResults=1, chunkSize=chunkSize)
        if failList:
            logger.info("model file failures (%d): %r", len(failList), failList)
        #
        for (modelFileIn, modelD, success) in resultList[0]:
            modelIdD = copy.deepcopy(modelD)
            if success:
                modelId = modelIdD.pop("modelId")
                mD[modelId] = modelIdD
            else:
                failD[modelFileIn] = modelIdD
        #
        logger.info("Completed with multi-proc status %r, failures %r, total models with data (%d)", ok, len(failList), len(mD))
        return mD, failD



##
# File:    ModelArchiveModelProvider.py
# Author:  Dennis Piehl
# Date:    18-Mar-2022
#
# Updates:
#
##
"""
Accessors for ModelArchive 3D In Silico Models (mmCIF).

"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

import datetime
import logging
import os.path
import time
from pathlib import Path
import glob

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.insilico3d.ModelReorganizer import ModelReorganizer

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger(__name__)


class ModelArchiveModelProvider:
    """Accessors for ModelArchive 3D in silico models (mmCIF)."""

    def __init__(self, cachePath, baseWorkPath=None, useCache=False, reload=True, **kwargs):
        """Initialize AlphaFoldModelProvider class.

        Args:
            cachePath (str): Path to directory where model files will be reorganized and stored permanently (also contains computed-model-cache.json file)  (i.e., 'computed-models'),
                             and will serve as the parent directory of where model files will be downloaded and where MA-specific cache file will sit (under 'work-dir/ModelArchive').
            useCache (bool, optional): Start from cache of already downloaded and/or reorganized model files. Defaults to False.
                                       When True, checks if last downloaded set of files is up-to-date and downloads any newly available models.
                                       When False (default), redownloads all model files.
            reload (bool, optional): Peform full reload (i.e., download/update) upon instantiation. Defaults to True.
        """
        # Use the same root cachePath for all types of insilico3D model sources, but with unique dirPath names (sub-directory)
        self.__cachePath = cachePath  # Cache path is where all model files will eventually be reorganized and stored in (i.e. "computed-models")
        self.__baseWorkPath = baseWorkPath if baseWorkPath else self.__cachePath
        self.__workPath = os.path.join(self.__baseWorkPath, "work-dir", "ModelArchive")  # Directory where model files will be downloaded (also contains MA-specific cache file)
        self.__dataSetCacheFile = os.path.join(self.__workPath, "model-download-cache.json")

        self.__mU = MarshalUtil(workPath=self.__workPath)
        self.__fU = FileUtil(workPath=self.__workPath)

        if reload:
            self.__oD, self.__createdDate = self.__reload(useCache=useCache, **kwargs)

    def testCache(self, minCount=0):  # Increase minCount once we are consistently downloading more than one data set
        if self.__oD and len(self.__oD) > minCount:
            return True
        else:
            return False

    def getCacheDirPath(self):
        return self.__workPath

    def reload(self, useCache, **kwargs):
        self.__oD, self.__createdDate = self.__reload(useCache=useCache, **kwargs)

    def __reload(self, **kwargs):
        """Reload cached list of ModelArchive model dataset files and check server for updated data sets,
        or re-download latest versions of all model data sets from server.

        Returns:
            oD (dict): dictionary of cached/downloaded model data sets with general metadata about each data set (local data path, size, # files, ...)
            createdDate (str): timestamp in isoformat of when the data cache was last created
        """
        try:
            oD = None
            createdDate = None
            ok = False
            startTime = time.time()
            startDateTime = datetime.datetime.now().isoformat()
            useCache = kwargs.get("useCache", True)
            baseUrl = kwargs.get("baseUrl", "https://www.modelarchive.org/api/projects")
            modelArchiveRequestedDatasetD = kwargs.get("modelArchiveRequestedDatasetD", {})
            if not modelArchiveRequestedDatasetD:  # Fill in default
                modelArchiveRequestedDatasetD = {
                    "ma-bak-cepc": {
                        "urlEnd": "ma-bak-cepc?type=materials_procedures__accompanying_data_file_name",
                        "fileName": "ma-bak-cepc.zip"
                    },
                }
            #
            self.__fU.mkdir(self.__workPath)
            #
            logger.info("useCache %r self.__dataSetCacheFile %r", useCache, self.__dataSetCacheFile)
            if useCache and self.__mU.exists(self.__dataSetCacheFile):
                logger.info("Loading data cache, %s.", self.__dataSetCacheFile)
                cacheD = self.__mU.doImport(self.__dataSetCacheFile, fmt="json")
                createdDate = cacheD["created"]
                oD = cacheD["data"]

                logger.info("Checking consistency of cached data with data available on server")
                for dataSet, pathD in modelArchiveRequestedDatasetD.items():
                    try:
                        cacheArchiveDir = oD[dataSet]["dataDirectory"]
                        cacheArchiveFileSize = oD[dataSet]["archiveFileSizeBytes"]
                        cacheArchiveFileDownloadDate = oD[dataSet]["lastDownloaded"]
                        cacheArchiveFileDownloadAge = (datetime.datetime.now() - datetime.datetime.fromisoformat(cacheArchiveFileDownloadDate)).days
                        if not os.path.exists(cacheArchiveDir):
                            logger.warning("Missing archive data for dataSet %s from server: %s", dataSet, pathD)
                        # If 120 days old, check consistency of local cache with data on ModelArchive
                        # (requires redownloading the entire archive file, since can't get requests.header info from the ModelArchive download URL)
                        if cacheArchiveFileDownloadAge > 120:
                            logger.info(
                                "Cached archive data for dataset %s last downloaded on %s (%d days ago): %s. Redownloading to check consistency with cached data.",
                                dataSet, cacheArchiveFileDownloadDate, cacheArchiveFileDownloadAge, pathD
                            )
                            dataSetFileName = pathD["fileName"]
                            dataSetFilePath = os.path.join(baseUrl, pathD["urlEnd"])
                            dataSetDataDumpDir = os.path.join(self.__workPath, dataSet.replace(" ", "_"))
                            self.__fU.mkdir(dataSetDataDumpDir)
                            dataSetFileDumpPath = os.path.join(dataSetDataDumpDir, dataSetFileName)
                            logger.info("Fetching file %s from server to local path %s", dataSetFilePath, dataSetFileDumpPath)
                            ok = self.__fU.get(dataSetFilePath, dataSetFileDumpPath)
                            logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
                            dataSetFileSize = int(self.__fU.size(dataSetFileDumpPath))
                            if cacheArchiveFileSize != dataSetFileSize:
                                logger.warning("Cached archive data for dataset %s not up-to-date with data archive on server: %s. You should Rebuild the cache!", dataSet, pathD)
                            logger.info("Deleting compressed dataset download, %s", dataSetFileDumpPath)
                            self.__fU.remove(dataSetFileDumpPath)
                        else:
                            logger.info(
                                "Cached archive data for dataset %s last downloaded on %s (%d days ago): %s. Skipping redownload of data.",
                                dataSet, cacheArchiveFileDownloadDate, cacheArchiveFileDownloadAge, pathD
                            )
                    except Exception as e:
                        logger.info("Failing on checking of cache data for dataSet %s: %s", dataSet, pathD)
                        logger.exception("Failing with %s", str(e))
            else:
                logger.info("Refetching all files from server.")
                cacheD = {}
                cacheD.update({"created": startDateTime, "data": {}})
                for dataSet, pathD in modelArchiveRequestedDatasetD.items():
                    try:
                        dataSetFileName = pathD["fileName"]
                        dataSetFilePath = os.path.join(baseUrl, pathD["urlEnd"])
                        dataSetDataDumpDir = os.path.join(self.__workPath, dataSet.replace(" ", "_"))
                        self.__fU.mkdir(dataSetDataDumpDir)
                        dataSetFileDumpPath = os.path.join(dataSetDataDumpDir, dataSetFileName)
                        #
                        logger.info("Fetching file %s from server to local path %s", dataSetFilePath, dataSetFileDumpPath)
                        ok = self.__fU.get(dataSetFilePath, dataSetFileDumpPath)
                        logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
                        dataSetFileSize = int(self.__fU.size(dataSetFileDumpPath))
                        ok = self.__fU.unbundleZipfile(dataSetFileDumpPath, dirPath=dataSetDataDumpDir)
                        logger.info("Completed unbundle (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
                        #
                        sD = {
                            "dataSetName": dataSet,
                            "archiveFileName": dataSetFileName,
                            "archiveFileSizeBytes": dataSetFileSize,
                            "lastDownloaded": startDateTime,
                            "dataDirectory": dataSetDataDumpDir,
                        }
                        #
                        logger.info("Clearing non-model files from extracted zip bundle...")
                        for nonModelFile in Path(dataSetDataDumpDir).glob("*.a3m"):
                            nonModelFile.unlink()
                        for nonModelFile in Path(dataSetDataDumpDir).glob("*_local_pairwise_qa.cif"):
                            nonModelFile.unlink()
                        #
                        if ok:
                            cacheD["data"].update({dataSet: sD})
                            self.__fU.remove(dataSetFileDumpPath)
                    #
                    except Exception as e:
                        logger.info("Failing on fetching and expansion of file for dataSet %s: %s", dataSet, pathD)
                        logger.exception("Failing with %s", str(e))

                createdDate = cacheD["created"]
                oD = cacheD["data"]
                ok = self.__mU.doExport(self.__dataSetCacheFile, cacheD, fmt="json", indent=3)
                logger.info("Export ModelArchive dataSet model data (%d) status %r", len(oD), ok)

        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return oD, createdDate

    def getArchiveDirList(self):
        archiveDirList = [self.__oD[k]["dataDirectory"] for k in self.__oD]

        return archiveDirList

    def getModelFileList(self, inputPathList=None):
        """Return a list of filepaths for all mmCIF models under the provided set of directories.

        Args:
            inputPathList (list): Optional (but recommended) list of paths to directories containing model mmCIF files to return.
                                  If not provided, method will retrieve all cached dataset paths, so all model files from all datasets will be returned in a single list;
                                  thus, it's recommended to provide a list of specific dataset directory to break the returned model list down into more manageable parts.

        Returns:
            (list): list of model mmCIF file paths (only matches compressed ".cif.gz" files, to ensure only one file per model)
        """

        if not inputPathList:
            inputPathList = self.getArchiveDirList()

        modelFileList = []

        for modelDir in inputPathList:
            try:
                modelFiles = glob.glob(os.path.join(modelDir, "*.cif"))
                modelFileList = [os.path.abspath(f) for f in modelFiles]
            except Exception as e:
                logger.exception("Failing with %s", str(e))

        return modelFileList

    def getArchiveDataDownloadDate(self):
        return self.__createdDate

    def getBaseDataPath(self):
        return self.__workPath

    def getArchiveDataCacheFilePath(self):
        return self.__dataSetCacheFile

    def getModelReorganizer(self, cachePath=None, useCache=True, workPath=None, **kwargs):
        cachePath = cachePath if cachePath else self.__cachePath
        workPath = workPath if workPath else self.__workPath
        return ModelReorganizer(cachePath=cachePath, useCache=useCache, workPath=workPath, **kwargs)

    def getComputedModelsDataPath(self):
        return self.__cachePath

    def reorganizeModelFiles(self, cachePath=None, useCache=True, inputModelList=None, **kwargs):
        """Reorganize model files from organism-wide model listing to hashed directory structure and rename files
        to follow internal identifier naming convention.

        Args:
            cachePath (str): Path to cache directory.
            inputModelList (list, optional): List of input model filepaths to reorganize; defaults to all models for all model datasets.
            **kwargs (optional):
                numProc (int): number of processes to use; default 2.
                chunkSize (int): incremental chunk size used for distribute work processes; default 20.
                keepSource (bool): whether to copy files to new directory (instead of moving them); default False.
                cacheFilePath (str): full filepath and name for cache file containing a dictionary of all reorganized models.

        Returns:
            (bool): True if successful; False otherwise.
        """
        try:
            ok = False
            #
            mR = self.getModelReorganizer(cachePath=cachePath, useCache=useCache, **kwargs)
            #
            if inputModelList:  # Only reorganize given list of model files
                ok = mR.reorganize(inputModelList=inputModelList, modelSource="ModelArchive", destBaseDir=self.__cachePath, useCache=useCache)
                if not ok:
                    logger.error("Reorganization of model files failed for inputModelList starting with item, %s", inputModelList[0])
            #
            else:  # Reorganize ALL model files for ALL available model datasets
                archiveDirList = self.getArchiveDirList()
                for archiveDir in archiveDirList:
                    inputModelList = self.getModelFileList(inputPathList=[archiveDir])
                    ok = mR.reorganize(inputModelList=inputModelList, modelSource="ModelArchive", destBaseDir=self.__cachePath, useCache=useCache)
                    if not ok:
                        logger.error("Reorganization of model files failed for dataset archive %s", archiveDir)
                        break
            return ok
        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            return False



##
# File:    ModelCacheProvider.py
# Author:  Dennis Piehl
# Date:    27-Apr-2022
#
# Updates:
#
# To Do:
#   - Need to make this able to retrieve from git stash or build locker (to get most recent set of models),
#     so that both coasts use the exact same set of data and cache (including populated timestamps)
# 21-Jul-2022  bv Allow for smaller number of test cases
##

"""
Provider class to retrieve the cache (dict) of downloaded model files and their associated metadata.

"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os.path
import os
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger(__name__)


class ModelCacheProvider(StashableBase):
    """Provider workflow class for retrieving and storing/reorganizing computed-model files from various sources."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        useCache = kwargs.get("useCache", True)
        self.__dirName = "computed-models"
        super(ModelCacheProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__holdingsRemotePath = kwargs.get("holdingsRemotePath", "http://computed-models-internal-west.rcsb.org/holdings/computed-models-holdings.json.gz")
        self.__fallbackUrl = "https://raw.githubusercontent.com/rcsb/py-rcsb_exdb_assets_stash/testing/stash/computed-models/computed-models-holdings.json.gz"
        self.__holdingsLocalPath = os.path.join(self.__dirPath, "computed-models-holdings.json.gz")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__mD, self.__idMapD = self.__reload(self.__dirPath, useCache)
        #

    def reload(self):
        self.__mD = self.__reload(self.__dirPath, True)
        return True

    def __reload(self, dirPath, useCache):
        """Reload cached list of downloaded computed-model files and associated metadata.

        Returns:
            (dict): dictionary of cached/downloaded computed-model data
        """
        startTime = time.time()
        ok = False
        #
        logger.info("useCache %r holdingsPath %r", useCache, self.__holdingsLocalPath)
        mD, idMapD = {}, {}
        try:
            if useCache and self.__mU.exists(self.__holdingsLocalPath):
                mD = self.__mU.doImport(self.__holdingsLocalPath, fmt="json")
            else:
                logger.info("Refetching computed-models holdings cache file from %r to %r", self.__holdingsRemotePath, self.__holdingsLocalPath)
                fU = FileUtil()
                fU.mkdir(dirPath)
                ok = fU.get(self.__holdingsRemotePath, self.__holdingsLocalPath)
                logger.info("Computed-model cache fetch status is %r", ok)
                if not ok:
                    logger.info("Refetching computed-models holdings cache fallback file from %r to %r", self.__holdingsRemotePath, self.__holdingsLocalPath)
                    ok = fU.get(self.__fallbackUrl, self.__holdingsLocalPath)
                    logger.info("Computed-model cache fallback fetch status is %r", ok)
                mD = self.__mU.doImport(self.__holdingsLocalPath, fmt="json")
            idMapD = self.getCompModelIdMap(modelCacheD=mD)
            if mD and idMapD:
                ok = True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)

        return mD, idMapD

    def testCache(self, minCount=1):
        logger.info("computed-model cache count %d", len(self.__mD))
        if self.__mD and len(self.__mD) > minCount:
            return True
        else:
            return False

    def getModelCacheDict(self):
        return self.__mD

    def getCompModelIdMap(self, modelCacheD=None):
        return self.__fetchCompModelIdMap(modelCacheD=modelCacheD)

    def __fetchCompModelIdMap(self, modelCacheD=None):
        """Get the ID mapping between the source model IDs and internal model identifiers for computational models.
        """
        #
        modelCacheD = modelCacheD if modelCacheD else self.__mD
        compModelIdMapD = {}
        try:
            for internalModelId, modelD in modelCacheD.items():
                compModelIdMapD.update({modelD["sourceId"]: internalModelId})
            logger.info("Computed-models mapped ID length: %d", len(compModelIdMapD))
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return compModelIdMapD

    def getInternalCompModelId(self, sourceId):
        """Get the mapped internal ID of computed model provided the original source ID.

        Args:
            sourceId (str): entry.id of computed model (e.g., "AF-P96541-F1")
        """
        compModelInternalId = None
        compModelIdMapD = self.__idMapD if self.__idMapD else self.getCompModelIdMap()
        #
        try:
            compModelInternalId = compModelIdMapD.get(sourceId, None)
            logger.debug("Computed-model sourceId (%s) mapped to internalId (%r)", sourceId, compModelInternalId)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return compModelInternalId

    def getCompModelData(self, compModelInternalId):
        """Get the associated metadata dict for a specific computed model.

        Args:
            compModelInternalId (str): internal computed-model identifier (e.g., "AF_AFP96541F1")
        """
        compModelD = {}
        try:
            compModelD = self.__mD.get(compModelInternalId, {})
            if not compModelD:
                logger.error("Unable to retrieve source URL for computed-model (%s)", compModelInternalId)
            logger.debug("Computed-model internalId (%s) modelData (%r)", compModelInternalId, compModelD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return compModelD



##
# File:    ModelProviderWorkflow.py
# Author:  Dennis Piehl
# Date:    6-Apr-2022
#
# Updates:
#
# To Do:
##

"""
Workflow wrapper for retrieving and reorganizing/storing computed models from various sources.

"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

import logging

from rcsb.utils.insilico3d.AlphaFoldModelProvider import AlphaFoldModelProvider
from rcsb.utils.insilico3d.ModelArchiveModelProvider import ModelArchiveModelProvider

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger(__name__)


class ModelProviderWorkflow:
    """Provider workflow class for retrieving and storing/reorganizing computed-model files from various sources."""

    def __init__(self, srcDir, destDir, modelProviders=None, useCache=True, numProc=4, chunkSize=40, **kwargs):
        """Initialize Model Provider Workflow instance.

        Args:
            srcDir (str): Directory path to which to download model files
            destDir (str): Path to directory where model files will be reorganized and stored permanently (also contains computed-model-cache.json file)
                           (i.e., path to 'computed-models'volume path). Note that this parameter corresponds to the "cachePath" parameter in the provider-specific classes.
            useCache (bool, optional): Start from cache of already downloaded and/or reorganized model files. Defaults to True.
                                       When True (default), checks if last downloaded set of files is up-to-date and downloads any newly available models.
                                       When False, redownloads all model files.
        """
        self.__srcDir = srcDir
        self.__destDir = destDir
        self.__useCache = useCache
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__modelProviders = modelProviders if modelProviders else ["AlphaFold", "ModelArchive"]
        #
        self.__alphaFoldRequestedSpeciesList = kwargs.get("alphaFoldRequestedSpeciesList", [])
        self.__modelArchiveRequestedDatasetD = kwargs.get("modelArchiveRequestedDatasetD", {})

        if "AlphaFold" in self.__modelProviders:
            self.__aFMP = AlphaFoldModelProvider(
                cachePath=self.__destDir,
                baseWorkPath=self.__srcDir,
                useCache=self.__useCache,
                reload=False,
                alphaFoldRequestedSpeciesList=self.__alphaFoldRequestedSpeciesList,
            )

        if "ModelArchive" in self.__modelProviders:
            self.__mAMP = ModelArchiveModelProvider(
                cachePath=self.__destDir,
                baseWorkPath=self.__srcDir,
                useCache=self.__useCache,
                reload=False,
                modelArchiveRequestedDatasetD=self.__modelArchiveRequestedDatasetD
            )

    def download(self, modelProviders=None, **kwargs):
        """Download model files from model provider sources.

        Args:
            modelProviders (list, optional): List of model providers to download from. Defaults to ["AlphaFold", "ModelArchive"].

        Returns:
            (bool): True if successful; False otherwise.
        """
        modelProviders = modelProviders if modelProviders else self.__modelProviders
        alphaFoldRequestedSpeciesList = kwargs.get("alphaFoldRequestedSpeciesList", self.__alphaFoldRequestedSpeciesList)
        modelArchiveRequestedDatasetD = kwargs.get("modelArchiveRequestedDatasetD", self.__modelArchiveRequestedDatasetD)

        ok = False
        try:
            for provider in modelProviders:
                if provider == "AlphaFold":
                    self.__aFMP.reload(useCache=self.__useCache, alphaFoldRequestedSpeciesList=alphaFoldRequestedSpeciesList)
                    ok = self.__aFMP.testCache()
                if provider == "ModelArchive":
                    self.__mAMP.reload(useCache=self.__useCache, modelArchiveRequestedDatasetD=modelArchiveRequestedDatasetD)
                    ok = self.__mAMP.testCache()
                if not ok:
                    logger.error("Problem downloading models and/or testing cache for provider, %s", provider)
                    return False
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def reorganize(self, modelProviders=None, keepSource=False, **kwargs):
        """Reorganize collection of downloaded model files into destination directory (e.g., in 'computed-models' volume path).

        Args:
            modelProviders (list, optional): List of model providers to update downloaded files from. Defaults to ["AlphaFold", "ModelArchive"].
            keepSource (bool, optional): whether to copy files to new directory (instead of moving them). Defaults to False.

        Returns:
            (bool): True if successful; False otherwise.
        """
        modelProviders = modelProviders if modelProviders else self.__modelProviders
        numProc = kwargs.get("numProc", self.__numProc)
        chunkSize = kwargs.get("chunkSize", self.__chunkSize)

        ok = False
        try:
            for provider in modelProviders:
                if provider == "AlphaFold":
                    ok = self.__aFMP.reorganizeModelFiles(useCache=self.__useCache, numProc=numProc, chunkSize=chunkSize, keepSource=keepSource)
                if provider == "ModelArchive":
                    ok = self.__mAMP.reorganizeModelFiles(useCache=self.__useCache, numProc=numProc, chunkSize=chunkSize, keepSource=keepSource)
                if not ok:
                    logger.info("Failed to reorganize models for provider, %s", provider)
                    return False
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    # def run(self, **kwargs):
    #     """Run the model provider workflow:
    #         1. Retrieve model files from all sources
    #         2. Reorganize and rename model files into computed-models data directory
    #     """
    #     return
